from ._colormap import ColorMap
from ._cubehelix import CubeHelixMap
from ._interpolated import InterpolatedMap
from ._listed import ListedMap
from ._palette_interpolated import PaletteInterpolatedMap
from ._segment_function import SegmentFunctionMap
from ._segment_interpolated import SegmentInterpolatedMap

__all__ = (
    "ColorMap",
    "CubeHelixMap",
    "InterpolatedMap",
    "ListedMap",
    "PaletteInterpolatedMap",
    "SegmentFunctionMap",
    "SegmentInterpolatedMap",
)
